"""adding PythonLambda

Revision ID: 6939fca644fb
Revises: 467af4a6e6b3
Create Date: 2019-03-14 20:39:59.813682

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6939fca644fb'
down_revision = '467af4a6e6b3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('lambdas', sa.Column('code', sa.Text()))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('lambdas', 'code')
    # ### end Alembic commands ###
